@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2002, 2005, 2007-2012 Free Software Foundation,
-   Inc.
+   Copyright (C) 1999-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -20,7 +19,7 @@
 int
 imap4d_bye (int reason)
 {
-  return imap4d_bye0 (reason, NULL);
+  return imap4d_bye_command (reason, NULL);
 }
 
 static jmp_buf pipejmp;
@@ -32,7 +31,7 @@ sigpipe (int sig)
 }
 
 int
-imap4d_bye0 (int reason, struct imap4d_command *command)
+imap4d_bye_command (int reason, struct imap4d_command *command)
 {
   int status = EX_SOFTWARE;
 
