@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2007, 2010-2012, 2014-2017 Free Software Foundation,
-   Inc.
+   Copyright (C) 2007-2019 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -37,6 +36,8 @@ int mu_assoc_install (mu_assoc_t assoc, const char *name, void *value);
 
 int mu_assoc_lookup_ref (mu_assoc_t assoc, const char *name, void *dataptr);
 int mu_assoc_install_ref (mu_assoc_t assoc, const char *name, void *pval);
+int mu_assoc_install_ref2 (mu_assoc_t assoc, const char *name,
+			   void *ret_val, const char **ret_name);
   
 int mu_assoc_get_iterator (mu_assoc_t assoc, mu_iterator_t *piterator);
 int mu_assoc_remove (mu_assoc_t assoc, const char *name);  
