@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2008, 2009 Free Software Foundation, Inc.
+   Copyright (C) 2008-2019 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
@@ -32,11 +32,12 @@
 #include <mailutils/message.h>
 #include <mailutils/observer.h>
 #include <mailutils/progmailer.h>
-#include <mailutils/vartab.h>
+#include <mailutils/wordsplit.h>
+#include <mailutils/io.h>
 
-#include <url0.h>
-#include <mailer0.h>
-#include <registrar0.h>
+#include <mailutils/sys/url.h>
+#include <mailutils/sys/mailer.h>
+#include <mailutils/sys/registrar.h>
 
 static int _url_prog_init     (mu_url_t);
 
@@ -44,6 +45,11 @@ static struct _mu_record _prog_record =
 {
   MU_PROG_PRIO,
   MU_PROG_SCHEME,
+  MU_RECORD_DEFAULT,
+  /* FIXME: MU_URL_USER could be used to request running with this
+     user privileges. */
+  MU_URL_SCHEME | MU_URL_PATH | MU_URL_QUERY, 
+  MU_URL_PATH,
   _url_prog_init,    /* url init.  */
   _mu_mailer_mailbox_init,  /* Mailbox entry.  */
   _mu_mailer_prog_init, /* Mailer entry.  */
@@ -68,11 +74,6 @@ _url_prog_uplevel (const mu_url_t orig, mu_url_t *up)
 static int
 _url_prog_init (mu_url_t url)
 {
-  /* not valid in a prog url */
-  if (url->secret || url->auth || url->host || url->port)
-    return EINVAL;
-  if (!url->path)
-    return EINVAL;
   url->_uplevel = _url_prog_uplevel;
   return 0;
 }
@@ -128,9 +129,8 @@ prog_open (mu_mailer_t mailer, int flags)
   if (access (path, X_OK) == -1)
     return errno;
 
-  mu_progmailer_set_debug (pm, mailer->debug);
   status = mu_progmailer_set_command (pm, path);
-  MU_DEBUG1 (mailer->debug, MU_DEBUG_TRACE, "prog (%s)\n", path);
+  mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_TRACE, ("prog (%s)", path));
   return status;
 }
 
@@ -140,26 +140,37 @@ prog_close (mu_mailer_t mailer)
   return mu_progmailer_close (mailer->data);
 }
 
-static int
-_expand_sender (const char *name, void *data, char **p)
-{
-  mu_address_t addr = data;
-  char *email;
-  int status = mu_address_aget_email (addr, 1, &email);
-
-  if (status != 0)
-    return status;
-  *p = email;
-  return 0;
-}
-
-struct ex_rcpt
+struct prog_exp
 {
   mu_message_t msg;
-  mu_address_t addr;
-  char *string;
+  mu_address_t sender_addr;
+  mu_address_t rcpt_addr;
 };
 
+static int
+_expand_err (char **ret, const char *prefix, int status)
+{
+  if (mu_asprintf (ret, "%s: %s", prefix, mu_strerror (status)))
+    return MU_WRDSE_NOSPACE;
+  return MU_WRDSE_USERERR;
+}
+
+static int
+_expand_sender (struct prog_exp *pe, char **ret)
+{
+  int status = mu_address_aget_email (pe->sender_addr, 1, ret);
+  switch (status)
+    {
+    case 0:
+      /* Even if ret==NULL, wordsplit will cope with it */
+      return MU_WRDSE_OK;
+
+    case ENOMEM:
+      return MU_WRDSE_NOSPACE;
+    }
+  return _expand_err (ret, "getting email", status);
+}
+
 static int
 address_add (mu_address_t *paddr, const char *value)
 {
@@ -207,74 +218,79 @@ message_read_rcpt (mu_message_t msg, mu_address_t *paddr)
 }
 
 static int
-_expand_rcpt (const char *name, void *data, char **p)
+_expand_rcpt (struct prog_exp *pe, char **ret)
 {
-  struct ex_rcpt *exrcpt = data;
   int status;
-
-  if (!exrcpt->string)
-    {
-      size_t i, count = 0;
-      size_t len = 0;
-      char *str;
-      mu_address_t tmp_addr = NULL, addr;
+  size_t i, count = 0;
+  size_t len = 0;
+  char *str;
+  mu_address_t tmp_addr = NULL, addr;
       
-      if (exrcpt->addr)
-	addr = exrcpt->addr;
-      else
-	{
-	  status = message_read_rcpt (exrcpt->msg, &tmp_addr);
-	  if (status)
-	    {
-	      mu_address_destroy (&tmp_addr);
-	      return status;
-	    }
-	  addr = tmp_addr;
-	}
-	    
-      mu_address_get_count (addr, &count);
-      for (i = 1; i <= count; i++)
-	{
-	  const char *email;
-	  if (i > 1)
-	    len++;
-	  if ((status = mu_address_sget_email (addr, i, &email)) != 0)
-	    {
-	      mu_address_destroy (&tmp_addr);
-	      return status;
-	    }
-	  len += strlen (email);
-	}
-
-      str = malloc (len + 1);
-      if (!str)
+  if (pe->rcpt_addr)
+    addr = pe->rcpt_addr;
+  else
+    {
+      status = message_read_rcpt (pe->msg, &tmp_addr);
+      if (status)
 	{
 	  mu_address_destroy (&tmp_addr);
-	  return ENOMEM;
+	  return _expand_err (ret, "reading recipients", status);
 	}
-      exrcpt->string = str;
-      
-      for (i = 1; i <= count; i++)
+      addr = tmp_addr;
+    }
+	    
+  mu_address_get_count (addr, &count);
+  for (i = 1; i <= count; i++)
+    {
+      const char *email;
+      if (i > 1)
+	len++;
+      if ((status = mu_address_sget_email (addr, i, &email)) != 0)
 	{
-	  const char *email;
-	  if (i > 1)
-	    *str++ = ' ';
-	  if (mu_address_sget_email (addr, i, &email))
-	    continue;
-	  strcpy (str, email);
-	  str += strlen (email);
+	  mu_address_destroy (&tmp_addr);
+	  return _expand_err (ret, "reading email", status);
 	}
-      *str = 0;
+      if (email)
+	len += strlen (email);
+    }
+
+  str = malloc (len + 1);
+  if (!str)
+    {
       mu_address_destroy (&tmp_addr);
-    }  
-  *p = exrcpt->string;
-  return 0;
+      return MU_WRDSE_NOSPACE;
+    }
+  *ret = str;
+      
+  for (i = 1; i <= count; i++)
+    {
+      const char *email;
+      if (i > 1)
+	*str++ = ' ';
+      if (mu_address_sget_email (addr, i, &email) || !email)
+	continue;
+      strcpy (str, email);
+      str += strlen (email);
+    }
+  *str = 0;
+  mu_address_destroy (&tmp_addr);
+  
+  return MU_WRDSE_OK;
 }
 
-void
-_free_rcpt (void *data, char *value)
+#define SEQ(s, n, l) \
+  (((l) == (sizeof(s) - 1)) && memcmp (s, n, l) == 0)
+
+static int
+prog_getvar (char **ret, const char *name, size_t nlen, void *data)
 {
-  free (value);
+  struct prog_exp *pe = data;
+  
+  if (SEQ ("sender", name, nlen))
+    return _expand_sender (pe, ret);
+  if (SEQ ("rcpt", name, nlen))
+    return _expand_rcpt (pe, ret);
+  return MU_WRDSE_UNDEF;
 }
 
 static int
@@ -283,41 +299,61 @@ url_to_argv (mu_url_t url, mu_message_t msg,
 	     int *pargc, char ***pargv)
 {
   int rc;
-  mu_vartab_t vtab;
-  struct ex_rcpt ex_rcpt;
+  struct prog_exp pe;
   char **query;
   size_t i;
   size_t argc;
   char **argv;
+  struct mu_wordsplit ws;
+  int wsflags;
   
-  ex_rcpt.msg = msg;
-  ex_rcpt.addr = to;
-  ex_rcpt.string = NULL;
-  mu_vartab_create (&vtab);
-  mu_vartab_define_exp (vtab, "sender", _expand_sender, NULL, from);
-  mu_vartab_define_exp (vtab, "rcpt", _expand_rcpt, _free_rcpt, &ex_rcpt);
+  pe.msg = msg;
+  pe.rcpt_addr = to;
+  pe.sender_addr = from;
+
+  ws.ws_getvar = prog_getvar;
+  ws.ws_closure = &pe;
+  wsflags = MU_WRDSF_NOSPLIT | MU_WRDSF_NOCMD |
+            MU_WRDSF_GETVAR | MU_WRDSF_CLOSURE;
 
   rc = mu_url_sget_query (url, &argc, &query);
   if (rc)
     return rc;
 
-  argv = calloc (argc + 1, sizeof (argv[0]));
+  argv = calloc (argc + 2, sizeof (argv[0]));
   if (!argv)
     return ENOMEM;
 
+  rc = mu_url_aget_path (url, &argv[0]);
+  if (rc)
+    {
+      free (argv);
+      return rc;
+    }
+  
   for (i = 0; i < argc; i++)
     {
-      if ((rc = mu_vartab_expand (vtab, query[i], &argv[i])))
+      if (mu_wordsplit (query[i], &ws, wsflags))
 	{
 	  mu_argcv_free (i, argv);
-	  mu_vartab_destroy (&vtab);
-	  return rc;
+	  mu_wordsplit_free (&ws);
+	  return errno;
 	}
+      if (ws.ws_wordc == 0)
+	argv[i+1] = strdup ("");
+      else
+	argv[i+1] = strdup (ws.ws_wordv[0]);
+      if (!argv[i+1])
+        {
+          mu_argcv_free(i, argv);
+          return ENOMEM;
+        }
+      wsflags |= MU_WRDSF_REUSE;
     }
-  argv[i] = NULL;
+  argv[i+1] = NULL;
+  if (wsflags & MU_WRDSF_REUSE)
+    mu_wordsplit_free (&ws);
   
-  mu_vartab_destroy (&vtab);
-
   *pargc = argc;
   *pargv = argv;
   return 0;
@@ -336,30 +372,29 @@ prog_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
   status = mu_url_sget_path (mailer->url, &command);
   if (status && status != MU_ERR_NOENT)
     {
-      MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		 "cannot get path from URL: %s\n",
-		 mu_strerror (status));
+      mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		("cannot get path from URL: %s",
+		 mu_strerror (status)));
       return status;
     }
   status = mu_progmailer_set_command (pm, command);
   if (status)
     {
-      MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		 "cannot set progmailer command: %s\n",
-		 mu_strerror (status));
+      mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		("cannot set progmailer command: %s",
+		 mu_strerror (status)));
       return status;
     }
       
   status = url_to_argv (mailer->url, msg, from, to, &argc, &argv);
   if (status)
     {
-      MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		 "cannot convert URL to command line: %s\n",
-		 mu_strerror (status));
+      mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		("cannot convert URL to command line: %s",
+		 mu_strerror (status)));
       return status;
     }
 
-  mu_progmailer_set_debug (pm, mailer->debug);
   status = mu_progmailer_open (pm, argv);
   if (status == 0)
     {
@@ -368,9 +403,8 @@ prog_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
 	mu_observable_notify (mailer->observable, MU_EVT_MAILER_MESSAGE_SENT,
 			      msg);
       else
-	MU_DEBUG1 (mailer->debug, MU_DEBUG_ERROR,
-		   "progmailer error: %s\n",
-		   mu_strerror (status));
+	mu_debug (MU_DEBCAT_MAILER, MU_DEBUG_ERROR,
+		  ("progmailer error: %s", mu_strerror (status)));
     }
   
   mu_argcv_free (argc, argv);
@@ -379,6 +413,6 @@ prog_send_message (mu_mailer_t mailer, mu_message_t msg, mu_address_t from,
 
 #else
 #include <stdio.h>
-#include <registrar0.h>
+#include <mailutils/sys/registrar.h>
 mu_record_t mu_prog_record = NULL;
 #endif
