@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2002, 2005, 2007, 2009-2012, 2014-2017 Free
-   Software Foundation, Inc.
+   Copyright (C) 1999-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -58,7 +57,7 @@ mail_print_msg (msgset_t *mspec, mu_message_t mesg, void *data)
   out = open_pager (lines);
 
   if (mailvar_is_true ("showenvelope"))
-    print_envelope (mspec, mesg, "From");
+    print_stream_envelope (out, mspec, mesg, "From");
   
   if (*(int *) data) /* print was called with a lowercase 'p' */
     {
