@@ -1,6 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2003, 2005-2012, 2014-2016 Free Software Foundation,
-   Inc.
+   Copyright (C) 2003-2019 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -221,7 +220,7 @@ elt_fixup (void *item, void *data)
     exit (1);
   elt->msg = mh_file_to_message (NULL, elt->file_name);
   if (!elt->msg)
-    return 1;
+    return MU_ERR_USER0;
 
   return 0;
 }
@@ -362,7 +361,7 @@ open_mailer ()
   status = mu_mailer_open (mailer, MU_STREAM_RDWR);
   if (status)
     {
-      mu_error (_("cannot open mailer `%s'"), url);
+      mu_error (_("cannot open mailer `%s': %s"), url, mu_strerror (status));
       return NULL;
     }
   return mailer;
@@ -562,9 +561,13 @@ backup_file (const char *file_name)
 
   if (unlink (new_name) && errno != ENOENT)
     mu_diag_funcall (MU_DIAG_ERROR, "unlink", new_name, errno);
-  else if (rename (file_name, new_name))
-    mu_error (_("cannot rename `%s' to `%s': %s"),
-	      file_name, new_name, mu_strerror (errno));
+  else
+    {
+      int rc = mu_rename_file (file_name, new_name, MU_COPY_OVERWRITE);
+      if (rc)
+	mu_error (_("cannot rename %s to %s: %s"),
+		  file_name, new_name, mu_strerror (errno));
+    }
   free (new_name);
 }
 
@@ -619,7 +622,7 @@ _action_send (void *item, void *data)
 	    mu_header_set_value (hdr, MU_HEADER_X_MAILER,
 				 DEFAULT_X_MAILER, 0);
 	  else if (strcmp (p, "no"))
-	    mu_header_set_value (hdr, MU_HEADER_X_MAILER, p, 0);
+	    mu_header_remove (hdr, MU_HEADER_X_MAILER, 1);
 	}
     }
   
@@ -629,7 +632,7 @@ _action_send (void *item, void *data)
   
   mailer = open_mailer ();
   if (!mailer)
-    return 1;
+    return MU_ERR_FAILURE;
 
   WATCH ((_("Sending message %s"), elt->file_name));
   if (split_message)
@@ -646,7 +649,7 @@ _action_send (void *item, void *data)
   if (rc)
     {
       mu_error(_("cannot send message: %s"), mu_strerror (rc));
-      return 1;
+      return MU_ERR_FAILURE;
     }
 
   WATCH ((_("Destroying the mailer")));
